/* filecache.hpp */
/*
Copyright (C) 2010, Igor Yudincev
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the authors may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <string>
#include <unordered_map>
#include "md5file.hpp"


class FileCacheItem {
public:
	FILETIME lastUpdate;
	std::string hash;    //MD5- 
	std::string hashLF;  //MD5-    CR
	bool bHashLF;        //  hashLF

	FileCacheItem() : lastUpdate(), hash(""), hashLF(""), bHashLF(false) {}
	FileCacheItem(FILETIME lastUpdate_, std::string hash_, std::string hashLF_, bool bHashLF_)
		: lastUpdate(lastUpdate_), hash(hash_), hashLF(hashLF_), bHashLF(bHashLF_) {}
	FileCacheItem& operator=(const FileCacheItem &r) {
		if (this!=&r) {
			lastUpdate = r.lastUpdate;
			hash = r.hash;
			hashLF = r.hashLF;
			bHashLF = r.bHashLF;
		}
		return *this;
	}
};

class FileCache {
private:
	__int64 MaxFileSize;
	std::unordered_map<std::string,FileCacheItem> items;
public:
	FileCache(__int64 MaxFileSize_) : MaxFileSize(MaxFileSize_) {}

	std::string getMD5(const wchar_t *filepath, bool bCutCR)
	{
		//get timestamp
		HANDLE h = CreateFile(filepath, GENERIC_READ,
			FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, 0);
		BY_HANDLE_FILE_INFORMATION info={0};
		GetFileInformationByHandle(h,&info);
		if (MaxFileSize!=0 &&
			((__int64)info.nFileSizeLow|(__int64)info.nFileSizeHigh<<32)>MaxFileSize)
		{
			CloseHandle(h);
			return "";
		}

		FileCacheItem *pItem = NULL;

		//convert filename to UTF-8
		int nUtf = WideCharToMultiByte(CP_UTF8, 0, filepath, -1, NULL, 0,
			NULL, NULL);
		char *pUtfName = new char[nUtf];
		if (WideCharToMultiByte(CP_UTF8, 0, filepath, -1, pUtfName, nUtf,
			NULL, NULL)!=0)
		{
			pItem = &items[pUtfName];
		}
		delete[] pUtfName;
		if (pItem==NULL) { //       
			CloseHandle(h);
			return "";
		}

		if (CompareFileTime(&info.ftLastWriteTime, &pItem->lastUpdate)!=0) {
			//  ,    (  bCutCR==true   )
			pItem->lastUpdate = info.ftLastWriteTime;
			pItem->hash = md5file(h, false);
			pItem->bHashLF = false;
		}
		else if (bCutCR && !pItem->bHashLF) { //     
			pItem->hashLF = md5file(h, true);
			pItem->bHashLF = true;
		}
		CloseHandle(h);
		return (bCutCR) ? pItem->hashLF : pItem->hash;
	}
};
