/* md5file.cpp */
/*
Copyright (C) 2010, Igor Yudincev
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the authors may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "openssl/md5.h"
#include "md5file.hpp"


#define HEXDIGIT(d) ( (d)>9 ? (d)+87 : (d)+'0' )


// return hex representation of digest as string
static std::string MD5hexdigest(const unsigned char *digest) {
	char buf[33];
	for (int i=0; i<16; i++) {
    	buf[i*2  ] = HEXDIGIT(digest[i]>>4);
    	buf[i*2+1] = HEXDIGIT(digest[i]&0x0F);
	}
	buf[32]=0;

	return std::string(buf);
}


std::string md5file(HANDLE hFile, bool bCutCR) {
	static const DWORD BUF_SIZE = 0x10000;
	char filebuf[BUF_SIZE];
	DWORD sizeOrig;
	MD5_CTX ctx;
	unsigned char digest[16];

	MD5_Init(&ctx);
	while (ReadFile(hFile, filebuf, BUF_SIZE, &sizeOrig, NULL) && sizeOrig!=0) {
		DWORD sizeNew = sizeOrig;
		if (bCutCR) { //  CR
			char *pPrevCR = (char*)memchr(filebuf, 13, sizeOrig);
			char *pWrite = pPrevCR;
			if (pPrevCR!=NULL) {
				//  .CR,    . .CR
				char *pCurrCR;
				while (sizeNew!=0 && (pCurrCR = (char*)memchr(pPrevCR+1, 13, sizeOrig-(pPrevCR+1-filebuf)))!=NULL) {
					if (pPrevCR+1 < pCurrCR) {
						size_t nWrite = pCurrCR-pPrevCR-1;
						memcpy(pWrite, pPrevCR+1, nWrite);
						pWrite += nWrite;
					}
					pPrevCR = pCurrCR;
					sizeNew--;
				}
				//     CR
				if (sizeNew!=0) {
					pCurrCR = filebuf+sizeOrig;
					if (pPrevCR+1 < pCurrCR)
						memcpy(pWrite, pPrevCR+1, pCurrCR-pPrevCR-1);
					sizeNew--;
				}
			}
		}
		if (sizeNew!=0)
			MD5_Update(&ctx, filebuf, sizeNew);
	}
	MD5_Final(digest, &ctx);
	return MD5hexdigest(digest);
}
