/* svn_c0.cpp */
/*
Copyright (C) 2010, Igor Yudincev
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the authors may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <shlwapi.h>
#include <string>
#include "Plugin.hpp"
#include "mystring.hpp"
#include "RegKey.hpp"
#include "entries.hpp"
#include "filecache.hpp"


#define ArraySize(a)  (sizeof(a)/sizeof(a[0]))

#define PluginName L"1svn_c0"
#define MaxFileSizeDef 0x80000

enum CompareLng {
	MNoLngStringDefined = -1,

	MOk,
	MCancel,

	MShowRevision,
	MShowModified,
	MCheckFileSize,
	MBytes,
};

struct InitDialogItem {
	int Type;
	int X1, Y1, X2, Y2;
	int Focus;
	DWORD_PTR Selected;
	int Flags;
	int DefaultButton;
	const TCHAR *Data;
};

static struct PluginStartupInfo info;
static EntriesFile *entries;
static FileCache *cache;

static struct Options {
	int bShowRevision;
	int bShowModified;
	int MaxFileSize;
} options;


BOOL APIENTRY DllMain( HANDLE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
                     )
{
    return TRUE;
}


#if defined(__GNUC__)

#ifdef __cplusplus
extern "C" {
#endif
	BOOL WINAPI DllMainCRTStartup(HANDLE hDll, DWORD dwReason, LPVOID lpReserved);
	int WINAPI GetMinFarVersionW(void);
	void WINAPI SetStartupInfoW(const struct PluginStartupInfo *psi);
	void WINAPI GetPluginInfoW(struct PluginInfo *pi);
	int WINAPI GetCustomDataW(const wchar_t *FilePath, wchar_t **CustomData);
	void WINAPI FreeCustomDataW(wchar_t *CustomData);
	void WINAPI ExitFARW();
	int WINAPI ConfigureW(int ItemNumber);
#ifdef __cplusplus
};
#endif

#endif


TCHAR *__fastcall GetMsg(int MsgId)
{
	return const_cast<TCHAR *>(info.GetMsg(info.ModuleNumber, MsgId));
}

void InitDialogItems(InitDialogItem *Init, FarDialogItem *Item, int ItemsNumber)
{
	for (int I=0; I<ItemsNumber; I++)
	{
		Item[I].Type=Init[I].Type;
		Item[I].X1=Init[I].X1;
		Item[I].Y1=Init[I].Y1;
		Item[I].X2=Init[I].X2;
		Item[I].Y2=Init[I].Y2;
		Item[I].Focus=Init[I].Focus;
		Item[I].Reserved=Init[I].Selected;
		Item[I].Flags=Init[I].Flags;
		Item[I].DefaultButton=Init[I].DefaultButton;

		if ((DWORD_PTR)Init[I].Data<300)
			Item[I].PtrData = GetMsg((unsigned)(DWORD_PTR)Init[I].Data);
		else
			Item[I].PtrData = Init[I].Data;

		Item[I].MaxLen = 0;
	}
}



int WINAPI GetMinFarVersionW(void)
{
    return MAKEFARVERSION(2, 0, 1588);
}

void WINAPI SetStartupInfoW(const struct PluginStartupInfo *psi)
{
	info = *psi;
	lstrcpy(PluginRootKey, info.RootKey);
	lstrcat(PluginRootKey, L"\\");
	lstrcat(PluginRootKey, PluginName);

	options.bShowRevision = GetRegKey(HKEY_CURRENT_USER, L"", L"ShowRevision", 1);
	options.bShowModified = GetRegKey(HKEY_CURRENT_USER, L"", L"ShowModified", 1);
	options.MaxFileSize   = GetRegKey(HKEY_CURRENT_USER, L"", L"MaxFileSize" , 0);

    entries = new EntriesFile;
    cache = new FileCache(options.MaxFileSize);
}

void WINAPI GetPluginInfoW(struct PluginInfo *pi)
{
	static const TCHAR *PluginConfigStrings[1] = {L"SVN_C0"};

	pi->Flags = PF_PRELOAD;
	pi->PluginConfigStrings=PluginConfigStrings;
	pi->PluginConfigStringsNumber=ArraySize(PluginConfigStrings);
}

int WINAPI GetCustomDataW(const wchar_t *FilePath, wchar_t **CustomData)
{
	int res = FALSE;
	*CustomData = NULL;

	const wchar_t *pFilename = wcsrchr(FilePath, '\\');
	if (pFilename==NULL) return FALSE;
	pFilename++;
	wchar_t *EntriesDir=NULL;
	Assign(&EntriesDir, FilePath, pFilename-FilePath);

	if (entries->read(EntriesDir)) {
		//convert filename to UTF-8
		int nUtf = WideCharToMultiByte(CP_UTF8, 0, pFilename, -1, NULL, 0,
			NULL, NULL);
		char *pUtfName = new char[nUtf];
		if (WideCharToMultiByte(CP_UTF8, 0, pFilename, -1, pUtfName, nUtf,
			NULL, NULL)!=0)
		{
			//search for the filename in entries
			const Entry *entry = entries->find(pUtfName);
			if (entry!=NULL) {
				wchar_t *pRevision;
				size_t nRevision;
				nRevision = (options.bShowRevision) ? entry->revision.length() : 0;
				if (options.bShowModified) {
					*CustomData = new wchar_t[(nRevision+2)*sizeof(wchar_t)];
					if (entry->hash.length()!=1) {
						std::string md5(cache->getMD5(FilePath, false));
						if (md5.empty()) {                         //too long file
							(*CustomData)[0] = L'~';
						}
						else if (md5.compare(entry->hash)==0) {    //unchanged file
							(*CustomData)[0] = L' ';
						}
						else {
							std::string md5LF(cache->getMD5(FilePath, true));
							(*CustomData)[0] = (md5LF.compare(entry->hash)==0) ? L' ' : L'*';;
						}
					}
					else
						(*CustomData)[0] = L' ';
					(*CustomData)[1] = 0;
					pRevision = *CustomData+1;
				}
				else {
					*CustomData = new wchar_t[(nRevision+1)*sizeof(wchar_t)];
					pRevision = *CustomData;
				}
				if (nRevision==0 ||
					MultiByteToWideChar(CP_ACP, 0, entry->revision.c_str(), -1,
						pRevision, nRevision+1)==0)
				{
					*pRevision = 0;
				}
				res = TRUE;
			}
		}
		delete[] pUtfName;
	}
	SetLength(&EntriesDir, 0);
	return res;
}

void WINAPI FreeCustomDataW(wchar_t *CustomData)
{
	if (CustomData)
		delete[] CustomData;
}

void WINAPI ExitFARW()
{
	delete cache;
	delete entries;
}


LONG_PTR WINAPI WndProc(
  HANDLE   hDlg,
  int      Msg,
  int      Param1,
  LONG_PTR Param2
)
{
	if (Msg==DN_BTNCLICK) {
		int bShowModified = (int)info.SendDlgMessage(hDlg, DM_GETCHECK, 2, 0);
		int bCheckFileSize = (int)info.SendDlgMessage(hDlg, DM_GETCHECK, 3, 0);
		switch (Param1) {
		case 2:
			info.SendDlgMessage(hDlg, DM_ENABLE, 3, bShowModified);
			//fall through
		case 3:
			info.SendDlgMessage(hDlg, DM_ENABLE, 4, bShowModified && bCheckFileSize);
			break;
		}
	}
	return info.DefDlgProc(hDlg, Msg, Param1, Param2);
}

int WINAPI ConfigureW(int ItemNumber)
{
	InitDialogItem initItems[]={
	/*  0 */ {DI_DOUBLEBOX,  3,  1, 42,  8, 0, 0,                    0,               0, L"SVN_C0"},
	/*  2 */ {DI_CHECKBOX,   5,  2, 40,  2, 0, 0,                    0,               0, GetMsg(MShowRevision)},
	/*  1 */ {DI_CHECKBOX,   5,  3, 40,  3, 1, 0,                    0,               0, GetMsg(MShowModified)},
	/*  3 */ {DI_CHECKBOX,   9,  4, 40,  4, 0, 0,                    0,               0, GetMsg(MCheckFileSize)},
	/*  4 */ {DI_FIXEDIT,    13, 5, 24,  5, 0, (int)L"999999999999", DIF_MASKEDIT,    0, L""},
	/*  5 */ {DI_TEXT,       26, 5, 40,  5, 0, 0,                    0,               0, GetMsg(MBytes)},
	/*  6 */ {DI_TEXT,       0,  6,  0,  6, 0, 0,                    DIF_SEPARATOR,   0, L""},
	/*  7 */ {DI_BUTTON,     0,  7,  0,  7, 0, 0,                    DIF_CENTERGROUP, 1, GetMsg(MOk)},
	/*  8 */ {DI_BUTTON,     0,  7,  0,  7, 0, 0,                    DIF_CENTERGROUP, 0, GetMsg(MCancel)}
	};

	int size=ArraySize(initItems);
	FarDialogItem *dialogItems = new FarDialogItem[size];
	InitDialogItems(initItems, dialogItems, size);

	dialogItems[1].Selected = options.bShowRevision;
	dialogItems[2].Selected = options.bShowModified;
	dialogItems[3].Selected = options.MaxFileSize!=0;
    TCHAR szBuf[16];
	wsprintf(szBuf, L"%d", options.MaxFileSize);
	if (!dialogItems[2].Selected) {
		dialogItems[3].Flags |= DIF_DISABLE;
	}
	if (!dialogItems[2].Selected || !dialogItems[3].Selected) {
		dialogItems[4].Flags |= DIF_DISABLE;
	}
	dialogItems[4].PtrData = szBuf;

	int res = 0;
	HANDLE hDlg = info.DialogInit(info.ModuleNumber, -1, -1, 46, 10, L"Configure",
		dialogItems, size, 0, 0, WndProc, 0);
	if (hDlg!=INVALID_HANDLE_VALUE) {
		int exitCode = info.DialogRun(hDlg);
		if (exitCode == 7) {
			options.bShowRevision = (int)info.SendDlgMessage(hDlg, DM_GETCHECK, 1, 0);
			options.bShowModified = (int)info.SendDlgMessage(hDlg, DM_GETCHECK, 2, 0);
			if ((int)info.SendDlgMessage(hDlg, DM_GETCHECK, 3, 0)!=0) {
				int oldMaxFileSize = options.MaxFileSize;
				options.MaxFileSize = wcstol((const TCHAR *)info.SendDlgMessage(hDlg, DM_GETCONSTTEXTPTR, 4, 0), NULL, 10);
				if (options.MaxFileSize!=oldMaxFileSize) {
					delete cache;
					cache = new FileCache(options.MaxFileSize);
				}
			}
			else
				options.MaxFileSize = 0;
			SetRegKey(HKEY_CURRENT_USER, L"", L"ShowRevision", options.bShowRevision);
			SetRegKey(HKEY_CURRENT_USER, L"", L"ShowModified", options.bShowModified);
			SetRegKey(HKEY_CURRENT_USER, L"", L"MaxFileSize" , options.MaxFileSize  );
			info.Control(PANEL_ACTIVE, FCTL_REDRAWPANEL, 0, 0);
			info.Control(PANEL_PASSIVE, FCTL_REDRAWPANEL, 0, 0);
			res = 1;
		}
		info.DialogFree(hDlg);
	}
	delete[] dialogItems;
	return res;
}
